﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.AI;
using Pathfinding;
using System.Text.RegularExpressions;

[RequireComponent(typeof(Camera))]
public class CameraCapture : MonoBehaviour
{

    public Camera Cam;
    public int Width = 1368;
    public int Height = 912;

    public int Size = 2048;

    private Texture2D sceneTarget;
    private Texture2D depthTarget;
    private Texture2D truthTarget;

    private const string PATH = "D:/data/test";

    private int n = 0;
    private int nm = 8000;

    private static float spare;
    private static bool isSpareReady = false;

    

    void Awake()
    {

    }

    // Start is called before the first frame update
    void Start()
    {
        sceneTarget = new Texture2D(Width, Height);
        depthTarget = new Texture2D(Width, Height);
        truthTarget = new Texture2D(Width, Height);

        n = GetFileCount();
        nm = n + nm;
    }

    int GetFileCount()
    {
        string sp = PATH + "/scene/";
        var imageFiles = System.IO.Directory.GetFiles(sp);
        return imageFiles.Length;
    }

    // Update is called once per frame
    void Update()
    {
        n += 1;

        if (n >= nm)
        {
#if UNITY_EDITOR
            UnityEditor.EditorApplication.isPlaying = false;
#else
            Application.Quit();
#endif
        }
    }

    void LateUpdate()
    {
        RenderTexture rt = new RenderTexture(Width, Height, 24);
        Cam.targetTexture = rt;
        Texture2D screenShot = new Texture2D(Width, Height, TextureFormat.RGB24, false);
        Cam.Render();
        RenderTexture.active = rt;
        sceneTarget.ReadPixels(new Rect(0, 0, Width, Height), 0, 0);

        Cam.targetTexture = null;
        RenderTexture.active = null; // JC: added to avoid errors
        Destroy(rt);

        RayTrace();
        SaveImages();
    }

    // Marsaglia polar method implementation, derived from java implementation at https://en.wikipedia.org/wiki/Marsaglia_polar_method
    float getGaussian(float mean, float stdDev)
    {
        if (isSpareReady)
        {
            isSpareReady = false;
            return spare * stdDev + mean;
        }
        float u, v, s;
        do
        {
            u = Random.value * 2 - 1;
            v = Random.value * 2 - 1;
            s = u * u + v * v;
        } while (s >= 1 || s == 0);
        float mul = Mathf.Sqrt(-2.0f * Mathf.Log(s) / s);
        spare = v * mul;
        isSpareReady = true;
        return mean + stdDev * u * mul;
    }

    void SaveImages()
    {
		string dp2 = PATH + "/depth2" + n + ".png"
        string dp = PATH + "/depth/" + n + ".png";
        string tp = PATH + "/truth/" + n + ".png";
        string sp = PATH + "/scene/" + n + ".png";
		ScreenCapture.CaptureScreenshot(sp2);
        System.IO.File.WriteAllBytes(dp, depthTarget.EncodeToPNG());
        System.IO.File.WriteAllBytes(tp, truthTarget.EncodeToPNG());
        System.IO.File.WriteAllBytes(sp, sceneTarget.EncodeToPNG());
    }

    void RayTrace()
    {
        for (int x = 0; x < Width; x += 1)
        {
            for (int y = 0; y < Height; y += 1)
            {
                Ray ray = Cam.ScreenPointToRay(new Vector3(x, y, 0));
                float dist = DistanceCalc(ray);

                depthTarget.SetPixel(x, y, TraceDepth(dist));
                truthTarget.SetPixel(x, y, TraceSegmentation(ray, dist));
            }
        }
    }

    private static float DistanceCalc(Ray ray)
    {
        float dist = 12000;
        if (Physics.Raycast(ray, out RaycastHit hit))
        {
            dist = Vector3.Distance(ray.origin, hit.point);

            dist = Mathf.Min(120, dist); // Cut off at 120
            dist = Mathf.Max(0, dist);
            // cm precision
            dist = dist * 100;
            dist = Mathf.Round(dist);
        }
        return dist;
    }

    private Color TraceSegmentation(Ray ray, float distance)
    {
        if (distance >= 12000)
        {
            return Color.black;
        }
        if (Physics.Raycast(ray, out RaycastHit hit))
        {
            int mask = 1 << NavMesh.GetAreaFromName("Unsafe");
            if (NavMesh.SamplePosition(hit.point, out NavMeshHit nhit2, 0.3f, NavMesh.AllAreas))
            {
                if (Mathf.Abs(hit.point.x - nhit2.position.x) < 0.02 && Mathf.Abs(hit.point.z - nhit2.position.z) < 0.02 && nhit2.position.y > 50)
                {
                    if (nhit2.mask == mask)
                    {
                        return Color.yellow;
                    }
                    return Color.blue;
                }
            }
        }
        return Color.red;
    }

    private static Color TraceDepth(float dist)
    {
        // normalize 
        dist = dist / 12000;
        // encode as 16byte   
        dist = dist * 65535;

        // Write 16 byte to r and g channel of png (convert to 1 channel float16 in python)
        byte r = (byte)Mathf.Floor(dist / 256);
        byte g = (byte)Mathf.RoundToInt(dist % 256);
        return new Color32(r, g, 0, 255);
    }
}